/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#ifndef APPROACHPLATE_H_
#define APPROACHPLATE_H_

/* This file is based off the data in the "Metafile_XML_Definitions.pdf" file
 * included with the NACO d-TPP Digital Terminal Procedures" product.
 */

#include <time.h>
#include <string>
#include <stdint.h>

/* We can't include DataStore.h due to circular dependencies, so we do a
 * forward declaration instead.
 */
class DataStore;

#define FRONTMATTERFILENAME "FRNTMATTER.PDF"

class ApproachPlate
{
public:
	ApproachPlate();
	virtual ~ApproachPlate();

	enum ChartCodeEnum {CODEINVALID,MIN,STAR,IAP,APD,DP,DPO};
	enum UserActionEnum {ACTIONINVALID,ADD,CHANGE,REMOVE,UNCHANGED};
	enum SectionEnum {SECTIONINVALID,TOMIN,ALTMIN,RADMIN,STARS,NORMAL};
	enum CivilEnum {CIVILINVALID,CIVIL,DMAAC,NGA,NGAHIGH};
	enum ImgTypeEnum {PDF,JPEG};

	/* "Set" Access Functions */
	void SetCycle(std::string Cycle);
	void SetEffective(time_t Effective);
	void SetExpire(time_t Expire);
	void SetStateID(std::string StateID);
	void SetStateName(std::string StateName);
	void SetCityName(std::string CityName);
	void SetVolume(std::string Volume);
	void SetAirportID(std::string AirportID);
	void SetAirportMilitary(std::string AirportMilitary);
	void SetAirportMilitary(bool AirportMilitary);
	void SetAirportIdent(std::string AirportIdent);
	void SetAirportICAOIdent(std::string AirportICAOIdent);
	void SetSeq(std::string Seq);
	void SetSeq(int Seq);
	void SetCode(std::string Code);
	void SetCode(ChartCodeEnum Code);
	void SetName(std::string Name);
	void SetUserAction(std::string UserAction);
	void SetUserAction(char UserAction);
	void SetUserAction(UserActionEnum UserAction);
	void SetImgName(std::string ImgName);
	void SetImgType(enum ImgTypeEnum);
	void SetCNFlag(std::string CNFlag);
	void SetCNFlag(bool CNFlag);
	void SetCNSection(std::string CNSection);
	void SetCNSection(char CNSection);
	void SetCNSection(SectionEnum CNSection);
	void SetCNPage(std::string CNPage);
	void SetCNPage(int CNPage);
	void SetBVSection(std::string BVSection);
	void SetBVSection(char BVSection);
	void SetBVSection(SectionEnum BVSection);
	void SetBVPage(std::string BVPage);
	void SetBVPage(int BVPage);
	void SetProcUID(std::string ProcUID);
	void SetProcUID(int ProcUID);
	void SetTwoColored(std::string TwoColored);
	void SetTwoColored(bool TwoColored);
	void SetCivil(std::string Civil);
	void SetCivil(char Civil);
	void SetCivil(CivilEnum Civil);
	void SetFAANFD15(std::string FAANFD15);
	void SetFAANFD18(std::string FAANFD18);
	void SetCopter(std::string Copter);
	void SetCopter(bool Copter);

	/* "Get" Access Functions */
	std::string GetCycle();
	time_t GetEffective();
	time_t GetExpire();
	std::string GetStateID();
	std::string GetStateName();
	std::string GetCityName();
	std::string GetVolume();
	std::string GetAirportID();
	bool GetAirportMilitary();
	std::string GetAirportIdent();
	std::string GetAirportICAOIdent();
	int GetSeq();
	ApproachPlate::ChartCodeEnum GetCode();
	std::string GetName();
	ApproachPlate::UserActionEnum GetUserAction();
	std::string GetImgName();
	enum ImgTypeEnum GetImgType();
	bool GetCNFlag();
	ApproachPlate::SectionEnum GetCNSection();
	int GetCNPage();
	ApproachPlate::SectionEnum GetBVSection();
	int GetBVPage();
	int GetProcUID();
	bool GetTwoColored();
	std::string GetFAANFD15();
	std::string GetFAANFD18();
	bool GetCopter();
	ApproachPlate::CivilEnum GetCivil();

	void Copy(ApproachPlate *Other);
	bool Matchi(std::string str);
	void View(DataStore *Data);

	bool MakeBinary(struct ApproachPlate_Binary_struct *BinaryData);
	void SetFromBinary(struct ApproachPlate_Binary_struct *BinaryData);

private:
	std::string Cycle;
	time_t Effective,Expire;
	std::string StateID;
	std::string StateName;
	std::string CityName;
	std::string Volume;
	std::string AirportID;
	bool AirportMilitary;
	std::string AirportIdent;
	std::string AirportICAOIdent;
	unsigned int Seq;
	ChartCodeEnum Code;
	std::string Name;
	UserActionEnum UserAction;
	std::string ImgName;
	enum ImgTypeEnum ImgType;
	bool CNFlag;
	SectionEnum CNSection;
	unsigned int CNPage;
	SectionEnum BVSection;
	unsigned int BVPage;
	unsigned int ProcUID;
	bool TwoColored;
	CivilEnum Civil;
	std::string FAANFD15;
	std::string FAANFD18;
	bool Copter;

	std::string toUpper(std::string str);
};

/* Increment the version when the structure changes */
#define APPROACHPLATE_BINARY_STRUCT_VER 1
#ifdef WIN32
#pragma pack(1)
#endif
struct ApproachPlate_Binary_struct {
	/* Don't store the Cycle, Effective, Expire, Volume, Airport ID,
	 * info with the approach - it should be stored in a larger structure
	 */
	char StateID[2];
	char StateName[20];
	char CityName[50];
	uint32_t Seq;
	uint8_t Code;
	char Name[50];
	uint8_t UserAction;
	char ImgName[40];
	uint8_t ImgType;
	bool CNFlag;
	uint8_t CNSection;
	uint16_t CNPage;
	uint8_t BVSection;
	uint16_t BVPage;
	uint32_t ProcUID;
	bool TwoColored;
	uint8_t Civil;
}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#endif /*APPROACHPLATE_H_*/
