/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <stdio.h>
#include <gdkmm/pixbuf.h>
#include <xtiffio.h>
#include <geotiff.h>
#include <geo_tiffp.h>
#include "SectionalChart.h"
#include "Util.h"

SectionalChart::SectionalChart(Config *Cfg, std::string FileName)
{
	uint32 h;

	/* Load GeoTiff data */
	geotif=NULL;
	chartraster=NULL;
	GeoReferenced=true;

	Util::ShellExpand(&FileName);
	this->FileName=FileName;

	tif=XTIFFOpen(FileName.c_str(), "r");

	if (tif) {
		double *scale;
		int result,count;
		result=TIFFGetField(tif, GTIFF_PIXELSCALE, &count, &scale);
		if (result) {
			XMetersPerPixel=scale[0];
			YMetersPerPixel=scale[0];
		} else {
			XMetersPerPixel=1.0;
			YMetersPerPixel=1.0;
		}

		/* Offset to the lower left corner of the image */
		EastingOffset=0.0;
		GetRasterHeight(&h);
		NorthingOffset=h;

		geotif=GTIFNew(tif);
		if (!GTIFGetDefn(geotif,&Defn))
			GeoReferenced=false;
		else if (!GTIFImageToPCS(geotif,&EastingOffset,&NorthingOffset))
			GeoReferenced=false;

		chartraster=NULL;
	} else {
		printf("Unable to open sectional chart file: %s\n",FileName.c_str());
	}
}

SectionalChart::~SectionalChart()
{
	if (geotif)
		GTIFFree(geotif);
	if (chartraster)
		_TIFFfree(chartraster);
	if (tif)
		XTIFFClose(tif);
}

bool SectionalChart::LoadRaster()
{
	uint32 w, h;

	if (!tif)
		return false;

	if (!GetRasterSize(&w,&h))
		return false;

	if (chartraster)
	{
		_TIFFfree(chartraster);
	}

	chartraster = (uint32*) _TIFFmalloc(w*h * sizeof (uint32));
	if (chartraster != NULL) {
		if (TIFFReadRGBAImageOriented(tif, w, h, chartraster, ORIENTATION_TOPLEFT, 0)) {
			/* Move the raw raster data into a pixbuf */
			pixbuf = Gdk::Pixbuf::create_from_data(
				(guint8*)chartraster,		// Data
				Gdk::COLORSPACE_RGB,		// Colorspace
				1,							// Has alpha
				8,							// Bits per sample
				w,							// Width
				h,							// Height
				w*4							// Rowstride (8-bit RGBA)
				);
		}
	}
	return true;
}

Glib::RefPtr<Gdk::Pixbuf> SectionalChart::GetPixbuf() {
	return pixbuf;
}

void SectionalChart::PixelToLonLat(int XPixel, int YPixel, double *Lon, double *Lat) {
	double XMeters,YMeters;
	PixelToMeters(XPixel,YPixel,&XMeters,&YMeters);
	MetersToLonLat(XMeters,YMeters,Lon,Lat);
}

void SectionalChart::PixelToMeters(int XPixel, int YPixel, double *XMeters, double *YMeters) {
	uint32 h;

	if (!GetRasterHeight(&h))
	{
		*XMeters=0.0;
		*YMeters=0.0;
		return;
	}

	/* Invert the Y axis */
	YPixel=h-YPixel-1;

	*XMeters=XPixel*XMetersPerPixel;
	*YMeters=YPixel*YMetersPerPixel;

	*XMeters+=EastingOffset;
	*YMeters+=NorthingOffset;
}

void SectionalChart::MetersToLonLat(double XMeters, double YMeters, double *Lon, double *Lat) {
	*Lon=XMeters;
	*Lat=YMeters;
	GTIFProj4ToLatLong(&Defn, 1, Lon, Lat);
}

void SectionalChart::LonLatToPixel(double Lon, double Lat, int *XPixel, int *YPixel) {
	double XMeters,YMeters;
	LonLatToMeters(Lon,Lat,&XMeters,&YMeters);
	MetersToPixel(XMeters,YMeters,XPixel,YPixel);
}

void SectionalChart::MetersToPixel(double XMeters, double YMeters, int *XPixel, int *YPixel) {
	uint32 h;

	if (!GetRasterHeight(&h))
	{
		*XPixel=0;
		*YPixel=0;
		return;
	}

	XMeters-=EastingOffset;
	YMeters-=NorthingOffset;

	*XPixel=(int)rint(XMeters/XMetersPerPixel);
	*YPixel=(int)rint(YMeters/YMetersPerPixel);

	/* Invert the Y axis */
	*YPixel=h-*YPixel-1;
}
void SectionalChart::LonLatToMeters(double Lon, double Lat, double *XMeters, double *YMeters) {
	*XMeters=Lon;
	*YMeters=Lat;
	GTIFProj4FromLatLong(&Defn, 1, XMeters, YMeters);
}

bool SectionalChart::GetValid()
{
	return (tif!=NULL);
}

/* The chart isn't "square" in Lon/Lat coordinates, so we need to scan to find the extremes.
 * We scan in fairly large steps to keep it fast, but that means we're not terribly accurate
 * either.
 */
bool SectionalChart::GetBounds(double *East, double *West, double *North, double *South)
{
	uint32 w,h,i;
	double Lon,Lat;
	if (tif)
	{
		if (!GetRasterSize(&w,&h))
			return false;
		*North=0.0;
		*South=90.0;
		*East=-180.0;
		*West=0.0;

		/* Scan the width */
		for (i=0;i<w-1000-1;i+=1000)
		{
			PixelToLonLat(i,0,&Lon,&Lat);
			if (Lat>*North)
				*North=Lat;
			PixelToLonLat(i,h-1,&Lon,&Lat);
			if (Lat<*South)
				*South=Lat;
		}

		/* Scan the height */
		for (i=0;i<h-1000-1;i+=1000)
		{
			PixelToLonLat(0,i,&Lon,&Lat);
			if (Lon<*West)
				*West=Lon;
			PixelToLonLat(w-1,i,&Lon,&Lat);
			if (Lon>*East)
				*East=Lon;
		}

	} else {
		return false;
	}

	return true;
}

bool SectionalChart::GetRasterSize(uint32 *w, uint32 *h)
{
	if (!tif)
		return false;

	if (!GetRasterWidth(w))
		return false;

	if (!GetRasterHeight(h))
		return false;

	return true;
}

bool SectionalChart::GetRasterWidth(uint32 *w)
{
	if (!tif)
		return false;

	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, w);
	return true;
}

bool SectionalChart::GetRasterHeight(uint32 *h)
{
	if (!tif)
		return false;

	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, h);
	return true;
}

bool SectionalChart::GetGeoReferenced()
{
	return GeoReferenced;
}

std::string SectionalChart::GetFileName()
{
	return FileName;
}
