/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <cstdlib>
#include <cerrno>
#include <cstring>

#include "Shapefile.h"
#include "Exception.h"
#include "Util.h"

ShapefilePolyLinePart::ShapefilePolyLinePart() {
}

ShapefilePolyLinePart::~ShapefilePolyLinePart() {
	Clear();
}

std::list<struct shapefile_point_struct*> *ShapefilePolyLinePart::GetPoints() {
	return &Points;
}

void ShapefilePolyLinePart::Clear() {
	std::list<struct shapefile_point_struct*>::iterator Pointi;

	for (Pointi=Points.begin();Pointi!=Points.end();Pointi++) {
		delete *Pointi;
	}
	Points.clear();
}

bool ShapefilePolyLinePart::ContainsPoint(double Lat, double Lon) {
	std::list<struct shapefile_point_struct*>::iterator Pointi, Previ, Nexti;
	static bool firsttime=true;
	static std::list<struct shapefile_point_struct*>::iterator LastMatchi, Begini;

	/* This function is used heavily to find duplicate points along state edges.  Since
	 * these duplicates are in sequence, we optimize by checking just before and after
	 * the last match first. */
	if (firsttime) {
		firsttime=false;
		LastMatchi=Points.begin();
		Begini=LastMatchi;
		LastMatchi++;	/* Start in a safe place */
	} else {
		if (Begini!=Points.begin()) {
			/* The Points element has changed, start over */
			LastMatchi=Points.begin();
			Begini=LastMatchi;
			LastMatchi++;	/* Start in a safe place */
		}
		Previ=LastMatchi;
		Nexti=LastMatchi;
		if (Previ!=Points.begin())
			Previ--;
		if ((*Previ)->x==Lon && (*Previ)->y==Lat) {
			LastMatchi=Previ;
			return true;
		}
		if (Nexti!=Points.end())
			Nexti++;
		if (Nexti!=Points.end()) {
			if ((*Nexti)->x==Lon && (*Nexti)->y==Lat) {
				LastMatchi=Nexti;
				return true;
			}
		}
	}

	/* The optimizations didn't find anything - go through the points by brute force */
	for (Pointi=Points.begin();Pointi!=Points.end();Pointi++) {
		if ((*Pointi)->x==Lon && (*Pointi)->y==Lat) {
			LastMatchi=Pointi;
			return true;
		}
	}

	return false;
}

ShapefilePolyLine::ShapefilePolyLine() {
}

ShapefilePolyLine::~ShapefilePolyLine() {
	Clear();
}

std::vector<ShapefilePolyLinePart*> *ShapefilePolyLine::GetParts() {
	return &Parts;
}

void ShapefilePolyLine::Clear() {
	std::vector<ShapefilePolyLinePart*>::iterator Parti;

	for (Parti=Parts.begin();Parti!=Parts.end();Parti++) {
		delete *Parti;
	}
	Parts.clear();
}

void ShapefilePolyLine::ResetBoundingBox() {
	unsigned int PointCount=0;
	/* Init to impossible extremes */
	BoundingBox[0]=400;
	BoundingBox[1]=400;
	BoundingBox[2]=-400;
	BoundingBox[3]=-400;

	/* Scan through all the points to find the extremes */
	for (std::vector<ShapefilePolyLinePart*>::iterator Parti=Parts.begin(); Parti!=Parts.end(); Parti++) {
		for (std::list<struct shapefile_point_struct*>::iterator Pointi=(*Parti)->GetPoints()->begin(); Pointi!=(*Parti)->GetPoints()->end(); Pointi++) {
			if ((*Pointi)->x<BoundingBox[1])
				BoundingBox[1]=(*Pointi)->x;
			if ((*Pointi)->y<BoundingBox[0])
				BoundingBox[0]=(*Pointi)->y;
			if ((*Pointi)->x>BoundingBox[3])
				BoundingBox[3]=(*Pointi)->x;
			if ((*Pointi)->y>BoundingBox[2])
				BoundingBox[2]=(*Pointi)->y;
			PointCount++;
		}
	}

	/* Handle the degenerate case with no data */
	if (PointCount==0) {
		BoundingBox[0]=0;
		BoundingBox[1]=0;
		BoundingBox[2]=0;
		BoundingBox[3]=0;
	}
}

void ShapefilePolyLine::SetBoundingBox(double LatMin, double LonMin, double LatMax, double LonMax) {
	BoundingBox[0]=LatMin;
	BoundingBox[1]=LonMin;
	BoundingBox[2]=LatMax;
	BoundingBox[3]=LonMax;
}

void ShapefilePolyLine::GetBoundingBox(double *LatMin, double *LonMin, double *LatMax, double *LonMax) {
	*LatMin=BoundingBox[0];
	*LonMin=BoundingBox[1];
	*LatMax=BoundingBox[2];
	*LonMax=BoundingBox[3];
}

bool ShapefilePolyLine::InBoundingBox(double Lat, double Lon) {
	return (Lat>=BoundingBox[0] && Lat<=BoundingBox[2] && Lon>=BoundingBox[1] && Lon<=BoundingBox[3]);
}

bool ShapefilePolyLine::ContainedIn(double LatMin, double LonMin, double LatMax, double LonMax) {
	double tmp;

	if (LatMin>LatMax) {
		tmp=LatMin;
		LatMin=LatMax;
		LatMax=tmp;
	}
	if (LonMin>LonMax) {
		tmp=LonMin;
		LonMin=LonMax;
		LonMax=tmp;
	}

	return !(LatMax<=BoundingBox[0] || LonMax<=BoundingBox[1] || LatMin>=BoundingBox[2] || LonMin>=BoundingBox[3]);
}

bool ShapefilePolyLine::ContainsPoint(double Lat, double Lon) {
	for (std::vector<ShapefilePolyLinePart*>::iterator Parti=Parts.begin(); Parti!=Parts.end(); Parti++) {
		if ((*Parti)->ContainsPoint(Lat,Lon)) {
			return true;
		}
	}
	return false;
}

int32_t ShapefilePolyLine::GetnPoints() {
	int32_t Count=0;

	for (std::vector<ShapefilePolyLinePart*>::iterator Parti=Parts.begin(); Parti!=Parts.end(); Parti++) {
		Count+=(*Parti)->GetPoints()->size();
	}
	return Count;
}

Shapefile::Shapefile() {
}

Shapefile::~Shapefile() {
	Clear();
}

std::vector<ShapefilePolyLine*> *Shapefile::GetPolyLines() {
	return &PolyLines;
}

std::vector<ShapefilePolygon*> *Shapefile::GetPolygons() {
	return &Polygons;
}

void Shapefile::GetPolyLineBounds(double *LatMin,double  *LonMin,double  *LatMax,double  *LonMax) {
	bool NoData=true;
	double BB[4];

	/* Initialize to impossible extremes */
	*LatMin=400;
	*LonMin=400;
	*LatMax=-400;
	*LonMax=-400;

	for (std::vector<ShapefilePolyLine*>::iterator Linei=PolyLines.begin(); Linei!=PolyLines.end(); Linei++) {
		if ((*Linei)->GetnPoints()!=0) {
			NoData=false;
			(*Linei)->GetBoundingBox(&(BB[1]),&(BB[0]),&(BB[3]),&(BB[2]));
			if (*LatMin>BB[1])
				*LatMin=BB[1];
			if (*LonMin>BB[0])
				*LonMin=BB[0];
			if (*LatMax<BB[3])
				*LatMax=BB[3];
			if (*LonMax<BB[2])
				*LonMax=BB[2];
		}
	}

	if (NoData) {
		*LatMin=0;
		*LonMin=0;
		*LatMax=0;
		*LonMax=0;
	}
}

void Shapefile::Clear() {
	for (std::vector<ShapefilePolyLine*>::iterator PolyLinei=PolyLines.begin();PolyLinei!=PolyLines.end();PolyLinei++) {
		delete *PolyLinei;
	}
	PolyLines.clear();
	for (std::vector<ShapefilePolygon*>::iterator Polygoni=Polygons.begin();Polygoni!=Polygons.end();Polygoni++) {
		delete *Polygoni;
	}
	Polygons.clear();
}

void Shapefile::Load(std::string Filename) {
	FILE *fp;
	struct shapefile_fileheader_struct fileheader;

	fp=fopen(Filename.c_str(),"rb");
	if (!fp) {
		throw Exception("Error opening shape file for reading.",Exception::LEVEL_ERROR);
	}

	if (fread(&fileheader,sizeof(fileheader),1,fp)!=1) {
		fclose(fp);
		throw Exception("Error reading shape file.",Exception::LEVEL_ERROR);
	}

	/* We're committed to load new data - clear any old data */
	Clear();

	switch (le32toh(fileheader.shapetype_l)) {
		case SHAPETYPE_POLYLINE:
			LoadPolyLines(fp);
			break;
		case SHAPETYPE_POLYGON:
			LoadPolygons(fp);
			break;
		default:
			fclose(fp);
			throw Exception("Unsupported shapes in shape file.",Exception::LEVEL_ERROR);
			break;
	}

	fclose(fp);
}

void Shapefile::LoadPolygons(FILE *fp) {
	struct shapefile_recordheader_struct recordheader;
	struct shapefile_record_struct record;
	size_t offset,result;
	int parti,pointi;
	void *newptr;
	ShapefilePolygon *NewPolygon;
	ShapefilePolyLinePart *NewPolyLinePart;

	while (fread(&recordheader,sizeof(recordheader),1,fp)>0) {
		offset=0;
		while (offset<be32toh(recordheader.contentlength_b)*2 && !feof(fp) && !ferror(fp)) {
			result=fread(&(record.shapetype_l),sizeof(record.shapetype_l),1,fp);
			if (result!=1) {
				fclose(fp);
				Clear();
				throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
			}

			offset+=sizeof(record.shapetype_l)*result;
			switch (record.shapetype_l) {
				case SHAPETYPE_NULL:
					break;
				case SHAPETYPE_POLYGON:
					result=fread(&(record.polygon.box),sizeof(record.polygon.box),1,fp);
					if (result!=1) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(record.polygon.box)*result;

					result=fread(&(record.polygon.numparts),sizeof(record.polygon.numparts),1,fp);
					if (result!=1) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(record.polygon.numparts)*result;

					result=fread(&(record.polygon.numpoints),sizeof(record.polygon.numpoints),1,fp);
					if (result!=1) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(record.polygon.numpoints)*result;

//					printf("Parts=%d Points=%d [%f,%f] [%f,%f]\n",record.polygon.numparts,record.polygon.numpoints,
//						record.polygon.box[1],record.polygon.box[0],
//						record.polygon.box[3],record.polygon.box[2]
//					);

					NewPolygon = new ShapefilePolygon();
					NewPolygon->SetBoundingBox(record.polygon.box[1],record.polygon.box[0],record.polygon.box[3],record.polygon.box[2]);

					/* Read parts */
					newptr=malloc(sizeof(int32_t)*record.polygon.numparts);
					if (!newptr) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error allocating memory (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					record.polygon.parts=(int32_t*)newptr;
					result=fread(record.polygon.parts,sizeof(int32_t),record.polygon.numparts,fp);
					if ((int32_t)result!=record.polygon.numparts) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(int32_t)*result;

					/* Read points */
					newptr=malloc(sizeof(struct shapefile_point_struct)*record.polygon.numpoints);
					if (!newptr) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error allocating memory (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					record.polygon.points=(struct shapefile_point_struct*)newptr;
					result=fread(record.polygon.points,sizeof(struct shapefile_point_struct),record.polygon.numpoints,fp);
					if ((int32_t)result!=record.polygon.numpoints) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(struct shapefile_point_struct)*result;

					/* Handle each part */
					for (parti=0;parti<record.polygon.numparts;parti++) {
						NewPolyLinePart=new ShapefilePolyLinePart();

						pointi=record.polygon.parts[parti];
						while (
							(parti+1<record.polygon.numparts && pointi<record.polygon.parts[parti+1]) ||
							(parti+1==record.polygon.numparts && pointi<record.polygon.numpoints)
							) {
							struct shapefile_point_struct *NewPoint=new struct shapefile_point_struct;
							NewPoint->x=record.polygon.points[pointi].x;
							NewPoint->y=record.polygon.points[pointi].y;
							if (record.polygon.points[pointi].x>180 || record.polygon.points[pointi].x<-180 ||
									record.polygon.points[pointi].y>180 || record.polygon.points[pointi].y<-180) {
								fclose(fp);
								free(record.polygon.points);
								throw Exception("Invalid data in shape file.",Exception::LEVEL_ERROR);
							}
							NewPolyLinePart->GetPoints()->push_back(NewPoint);
							pointi++;
						}
						NewPolygon->GetParts()->push_back(NewPolyLinePart);
					}
					Polygons.push_back(NewPolygon);

					free(record.polygon.points);
					free(record.polygon.parts);
					break;
				default:
					fclose(fp);
					throw Exception("Error Unexpected shape type in shape file.",Exception::LEVEL_ERROR);
					return;
			}
		}
	}

	/* Convert the polygons from the shapefile into raw polylines */
	std::vector<ShapefilePolyLine*> RawPolyLines;
	for (std::vector<ShapefilePolygon*>::iterator Polygoni=Polygons.begin(); Polygoni!=Polygons.end(); Polygoni++) {
		ShapefilePolyLine *NewPolyLine=new ShapefilePolyLine();
		for (std::vector<ShapefilePolyLinePart*>::iterator Parti=(*Polygoni)->GetParts()->begin(); Parti!=(*Polygoni)->GetParts()->end(); Parti++) {
			NewPolyLine->GetParts()->push_back(*Parti);
		}
		double LatMin,LonMin,LatMax,LonMax;
		(*Polygoni)->GetBoundingBox(&LatMin,&LonMin,&LatMax,&LonMax);
		NewPolyLine->SetBoundingBox(LatMin,LonMin,LatMax,LonMax);
		RawPolyLines.push_back(NewPolyLine);
	}
	/* Get rid of the polygons */
	Polygons.clear();

	/* At this point there are polygons or parts of polygons for each state.  As each state has a complete outline, bordering states
	 * have duplicate points.  Go through the list to cull these duplicates.
	 */
	double BB[4];
	for (std::vector<ShapefilePolyLine*>::iterator Linei=RawPolyLines.begin();Linei!=RawPolyLines.end();Linei++) {
		ShapefilePolyLine *NewLine=new ShapefilePolyLine();
		std::vector<ShapefilePolyLine*>::iterator SubLinei=Linei;
		std::vector<ShapefilePolyLine*> Neighbors;
		SubLinei++;
		/* Search through each remaining line to find potential neighbors to check for duplicate points */
		for (;SubLinei!=RawPolyLines.end();SubLinei++) {
			(*SubLinei)->GetBoundingBox(&(BB[0]),&(BB[1]),&(BB[2]),&(BB[3]));
			if ((*Linei)->ContainedIn(BB[0],BB[1],BB[2],BB[3])) {
				Neighbors.push_back(*SubLinei);
			}
		}
//This is a temporary hack for quick debugging
//if (Neighbors.size()<1)
		for (std::vector<ShapefilePolyLinePart*>::iterator Parti=(*Linei)->GetParts()->begin();Parti!=(*Linei)->GetParts()->end();Parti++) {
			ShapefilePolyLinePart* NewPart=new ShapefilePolyLinePart();
			bool Contig=true;

			for (std::list<struct shapefile_point_struct*>::iterator Pointi=(*Parti)->GetPoints()->begin(); Pointi!=(*Parti)->GetPoints()->end(); Pointi++) {
				bool Dupe=false;

				if (Pointi!=(*Parti)->GetPoints()->begin() && Pointi!=(*Parti)->GetPoints()->end()) {
					/* Find duplicates for any points except the first and last */
					for (std::vector<ShapefilePolyLine*>::iterator Neighbori=Neighbors.begin();Neighbori!=Neighbors.end();Neighbori++) {
						if ((*Neighbori)->InBoundingBox((*Pointi)->y,(*Pointi)->x) && (*Neighbori)->ContainsPoint((*Pointi)->y,(*Pointi)->x)) {
							Dupe=true;
							Contig=false;
							break;
						}
					}
				}
				if (!Dupe) {
					if (Contig) {
						NewPart->GetPoints()->push_back(*Pointi);
					} else {
						/* We're starting a new line segment that is not duplicated */
						NewLine->GetParts()->push_back(NewPart);
						PolyLines.push_back(NewLine);
						NewLine=new ShapefilePolyLine();
						NewPart=new ShapefilePolyLinePart();
						NewPart->GetPoints()->push_back(*Pointi);
						Contig=true;
					}
				}
			}
			NewLine->GetParts()->push_back(NewPart);
		}
		PolyLines.push_back(NewLine);
	}

	/* Remove any empty parts (all parts of the line were handled by its neighbors) */
	bool Done;
	for (std::vector<ShapefilePolyLine*>::iterator Linei=PolyLines.begin();Linei!=PolyLines.end();Linei++) {
		Done=false;
		while (!Done) {
			Done=true;
			for (std::vector<ShapefilePolyLinePart*>::iterator Parti=(*Linei)->GetParts()->begin();Parti!=(*Linei)->GetParts()->end();Parti++) {
				if ((*Parti)->GetPoints()->size()==0) {
					Done=false;
					(*Linei)->GetParts()->erase(Parti);
					break;
				}
			}
		}
	}

	/* Remove any empty lines */
	Done=false;
	while (!Done) {
		for (std::vector<ShapefilePolyLine*>::iterator Linei=PolyLines.begin();Linei!=PolyLines.end();Linei++) {
			Done=true;
			if ((*Linei)->GetParts()->size()==0) {
				Done=false;
				PolyLines.erase(Linei);
				break;
			}
		}
	}

	/* Generate the bounding boxes for the new polylines */
	for (std::vector<ShapefilePolyLine*>::iterator Linei=PolyLines.begin();Linei!=PolyLines.end();Linei++) {
		(*Linei)->ResetBoundingBox();
	}
}

void Shapefile::LoadPolyLines(FILE *fp) {
	struct shapefile_recordheader_struct recordheader;
	struct shapefile_record_struct record;
	size_t offset,result;
	int parti,pointi;
	void *newptr;
	ShapefilePolyLine *NewPolyLine;
	ShapefilePolyLinePart *NewPolyLinePart;

	while (fread(&recordheader,sizeof(recordheader),1,fp)>0) {
		offset=0;
		while (offset<be32toh(recordheader.contentlength_b)*2 && !feof(fp) && !ferror(fp)) {
			result=fread(&(record.shapetype_l),sizeof(record.shapetype_l),1,fp);
			if (result!=1) {
				fclose(fp);
				Clear();
				throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
			}

			offset+=sizeof(record.shapetype_l)*result;

			switch (le32toh(record.shapetype_l)) {
				case SHAPETYPE_NULL:
					break;
				case SHAPETYPE_POLYLINE:
					result=fread(&(record.polyline.box),sizeof(record.polyline.box),1,fp);
					if (result!=1) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(record.polyline.box)*result;

					result=fread(&(record.polyline.numparts),sizeof(record.polyline.numparts),1,fp);
					if (result!=1) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(record.polyline.numparts)*result;

					result=fread(&(record.polyline.numpoints),sizeof(record.polyline.numpoints),1,fp);
					if (result!=1) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(record.polyline.numpoints)*result;

//					printf("Parts=%d Points=%d [%f,%f] [%f,%f]\n",record.polyline.numparts,record.polyline.numpoints,
//						record.polyline.box[1],record.polyline.box[0],
//						record.polyline.box[3],record.polyline.box[2]
//					);

					NewPolyLine = new ShapefilePolyLine();
					NewPolyLine->SetBoundingBox(record.polyline.box[1],record.polyline.box[0],record.polyline.box[3],record.polyline.box[2]);

					/* Read parts */
					newptr=malloc(sizeof(int32_t)*record.polyline.numparts);
					if (!newptr) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error allocating memory (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					record.polyline.parts=(int32_t*)newptr;
					result=fread(record.polyline.parts,sizeof(int32_t),record.polyline.numparts,fp);
					if ((int32_t)result!=record.polyline.numparts) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(int32_t)*result;

					/* Read points */
					newptr=malloc(sizeof(struct shapefile_point_struct)*record.polyline.numpoints);
					if (!newptr) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error allocating memory (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					record.polyline.points=(struct shapefile_point_struct*)newptr;
					result=fread(record.polyline.points,sizeof(struct shapefile_point_struct),record.polyline.numpoints,fp);
					if ((int32_t)result!=record.polyline.numpoints) {
						fclose(fp);
						Clear();
						throw Exception(std::string("Error reading shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
					}
					offset+=sizeof(struct shapefile_point_struct)*result;

					/* Handle each part */
					for (parti=0;parti<record.polyline.numparts;parti++) {
						NewPolyLinePart=new ShapefilePolyLinePart();

						pointi=record.polyline.parts[parti];
						while (
							(parti+1<record.polyline.numparts && pointi<record.polyline.parts[parti+1]) ||
							(parti+1==record.polyline.numparts && pointi<record.polyline.numpoints)
							) {
							struct shapefile_point_struct *NewPoint=new struct shapefile_point_struct;
							NewPoint->x=record.polyline.points[pointi].x;
							NewPoint->y=record.polyline.points[pointi].y;
							if (record.polyline.points[pointi].x>180 || record.polyline.points[pointi].x<-180 ||
									record.polyline.points[pointi].y>180 || record.polyline.points[pointi].y<-180) {
								fclose(fp);
								free(record.polyline.points);
								throw Exception("Invalid data in shape file.",Exception::LEVEL_ERROR);
							}
							NewPolyLinePart->GetPoints()->push_back(NewPoint);
							pointi++;
						}
						NewPolyLine->GetParts()->push_back(NewPolyLinePart);
					}
					PolyLines.push_back(NewPolyLine);

					free(record.polyline.points);
					free(record.polyline.parts);
					break;
				default:
					fclose(fp);
					throw Exception("Error Unexpected shape type in shape file.",Exception::LEVEL_ERROR);
					return;
			}
		}
	}
}

void Shapefile::SavePolyLines(std::string Filename) {
	FILE *filefp,*indexfp;
	struct shapefile_fileheader_struct fileheader;
	struct shapefile_indexfileheader_struct indexheader;
	struct shapefile_recordheader_struct recordheader;
	struct shapefile_recordindex_struct recordindex;
	struct shapefile_record_struct record;
	int32_t nRecord, FileOffset=0;

	filefp=fopen((Filename+std::string(".shp")).c_str(),"wb");
	if (!filefp) {
		throw Exception(std::string("Error opening shape file for writing (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
	}

	indexfp=fopen((Filename+std::string(".shx")).c_str(),"wb");
	if (!indexfp) {
		fclose(filefp);
		throw Exception(std::string("Error opening shape file for writing (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
	}

	/* Build the file header */
	fileheader.filecode_b=htobe32(SHAPEFILE_FILECODE);
	fileheader.unused_b[0]=fileheader.unused_b[1]=fileheader.unused_b[2]=fileheader.unused_b[3]=fileheader.unused_b[4]=0;

	fileheader.filelength_b=sizeof(fileheader);
	for (std::vector<ShapefilePolyLine*>::iterator Linei=PolyLines.begin(); Linei!=PolyLines.end(); Linei++) {
		fileheader.filelength_b+=
			sizeof(recordheader)+
			sizeof(record.shapetype_l)+
			sizeof(record.polyline)-
			sizeof(record.polyline.parts)-
			sizeof(record.polyline.points)+
			(*Linei)->GetParts()->size()*sizeof(int32_t)+
			(*Linei)->GetnPoints()*sizeof(struct shapefile_point_struct);
	}
	fileheader.filelength_b=htobe32(fileheader.filelength_b/2);

	fileheader.version_l=htole32(SHAPEFILE_VERSION);
	fileheader.shapetype_l=htole32(SHAPETYPE_POLYLINE);
	GetPolyLineBounds(&(fileheader.ymin),&(fileheader.xmin),&(fileheader.ymax),&(fileheader.xmax));
	fileheader.zmin=0.0;
	fileheader.zmax=0.0;
	fileheader.mmin=0.0;
	fileheader.mmax=0.0;

	/* Save the file header */
	if (fwrite(&fileheader,sizeof(fileheader),1,filefp)!=1) {
		fclose(filefp);
		fclose(indexfp);
		throw Exception(std::string("Error writing shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
	}
	FileOffset+=sizeof(fileheader);

	/* Build the index header */
	memcpy(&indexheader,&fileheader,sizeof(indexheader));
	indexheader.filelength_b=htobe32(((sizeof(indexheader)+sizeof(recordheader)*PolyLines.size())/2));

	/* Save the index file header */
	if (fwrite(&indexheader,sizeof(indexheader),1,indexfp)!=1) {
		fclose(filefp);
		fclose(indexfp);
		throw Exception(std::string("Error writing shape file index (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
	}

	nRecord=1;	/* The record count starts from 1, not zero */
	for (std::vector<ShapefilePolyLine*>::iterator Linei=PolyLines.begin(); Linei!=PolyLines.end(); Linei++) {
		/* Build the record header for this line */
		recordheader.recordnumber_b=htobe32(nRecord);
		nRecord+=1;
		recordheader.contentlength_b=htobe32(
			(
				(
				sizeof(record.shapetype_l)+
				sizeof(record.polyline)-
				sizeof(record.polyline.parts)-
				sizeof(record.polyline.points)+
				(*Linei)->GetParts()->size()*sizeof(int32_t)+
				(*Linei)->GetnPoints()*sizeof(struct shapefile_point_struct)
				)/2
			)
		);
		/* Save the record header to the main shapefile */
		if (fwrite(&recordheader,sizeof(recordheader),1,filefp)!=1) {
			fclose(filefp);
			fclose(indexfp);
			throw Exception(std::string("Error writing shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
		}

		/* Save the record index */
		recordindex.offset_b=htobe32(FileOffset/2);
		recordindex.contentlength_b=recordheader.contentlength_b;
		if (fwrite(&recordindex,sizeof(recordindex),1,indexfp)!=1) {
			fclose(filefp);
			fclose(indexfp);
			throw Exception(std::string("Error writing shape index file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
		}

		FileOffset+=sizeof(recordheader);

		record.shapetype_l=htole32(SHAPETYPE_POLYLINE);
		(*Linei)->GetBoundingBox(&(record.polyline.box[1]),&(record.polyline.box[0]),&(record.polyline.box[3]),&(record.polyline.box[2]));
		record.polyline.numparts=(*Linei)->GetParts()->size();
		record.polyline.numpoints=(*Linei)->GetnPoints();
		if (fwrite(&record,sizeof(record.shapetype_l)+sizeof(record.polyline)-sizeof(record.polyline.parts)-sizeof(record.polyline.points),1,filefp)!=1) {
			fclose(filefp);
			fclose(indexfp);
			throw Exception(std::string("Error writing shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
		}
		FileOffset+=sizeof(record.shapetype_l)+sizeof(record.polyline)-sizeof(record.polyline.parts)-sizeof(record.polyline.points);

		/* Handle parts array */
		record.polyline.parts=new int32_t[record.polyline.numparts];
		int32_t PartOffset=0;
		int32_t i=0;
		for (std::vector<ShapefilePolyLinePart*>::iterator Parti=(*Linei)->GetParts()->begin(); Parti!=(*Linei)->GetParts()->end(); Parti++) {
			record.polyline.parts[i++]=PartOffset;
			PartOffset+=(*Parti)->GetPoints()->size();
		}

		if (fwrite(record.polyline.parts,sizeof(int32_t)*record.polyline.numparts,1,filefp)!=1) {
			fclose(filefp);
			fclose(indexfp);
			throw Exception(std::string("Error writing shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
		}
		delete record.polyline.parts;
		FileOffset+=sizeof(int32_t)*record.polyline.numparts;

		/* Handle points array */
		if (record.polyline.numpoints>0) {
			record.polyline.points=new struct shapefile_point_struct[record.polyline.numpoints];
			i=0;
			for (std::vector<ShapefilePolyLinePart*>::iterator Parti=(*Linei)->GetParts()->begin(); Parti!=(*Linei)->GetParts()->end(); Parti++) {
				for (std::list<struct shapefile_point_struct*>::iterator Pointi=(*Parti)->GetPoints()->begin(); Pointi!=(*Parti)->GetPoints()->end(); Pointi++) {
					record.polyline.points[i].x=(*Pointi)->x;
					record.polyline.points[i++].y=(*Pointi)->y;
				}
			}
			if (fwrite(record.polyline.points,sizeof(struct shapefile_point_struct)*record.polyline.numpoints,1,filefp)!=1) {
				fclose(filefp);
				fclose(indexfp);
				throw Exception(std::string("Error writing shape file (")+std::string(strerror(errno))+std::string(")."),Exception::LEVEL_ERROR);
			}
			delete record.polyline.points;
			FileOffset+=sizeof(struct shapefile_point_struct)*record.polyline.numpoints;
		}
	}

	fclose(filefp);
	fclose(indexfp);
}
